/*----------------------------------------------------------------------
	This is part of the PureLS Source code.

	Copyright (C) 2000-2003 The PureLS Development Team

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
	USA
----------------------------------------------------------------------*/


/*----------------------------------------------------------------------
 Lets just make sure we aren't depending on IE
----------------------------------------------------------------------*/
#define _WIN32_IE 0x0000


/*---System Includes--------------------------------------------------*/
#include <tchar.h>
#include <windows.h>
#include <stdio.h>
#include <ctype.h>
/*---Shell Includes---------------------------------------------------*/
#include "../sdk/include/plsdefs.h"
#include "plsapi.h"
/*--------------------------------------------------------------------*/



/*--------PL_StripWS----------------------------------------------------
 Purpose:
 - Strip white space from a string
 Note:
 - If the entire string is white space, it remains unmodified and the
   function returns FALSE.
----------------------------------------------------------------------*/
BOOL WINAPI PL_StripWS(
	LPTSTR buffer
)
{
	size_t length = (NULL != buffer) ? _tcslen(buffer):0;

	if( 0 != length )
	{
		LPTSTR tmpBuf = buffer;

		/*--------------------------------------------------------------
		 Remove white space in the front.
		--------------------------------------------------------------*/
		while( 0 != length && _istspace(*tmpBuf) )
		{
			++tmpBuf;
			--length;
		}

		if( 0 != length )
		{
			/*----------------------------------------------------------
			 Remove white space at the back
			----------------------------------------------------------*/
			while( 0 != length && _istspace(tmpBuf[length-1]) )
			{
				tmpBuf[--length] = 0;
			}

			/*----------------------------------------------------------
			 If we removed white space in the front, we need to move the
			 text to the front of the buffer.  We wait until after we
			 remove the trailing whitespace to do this, because it will
			 be slightly faster with less characters to copy.

			 NOTE: We shouldn't be using strcpy, because according to
			 MSDN, its behavior is undefined when src and dest buffers
			 overlap, which they do in this case.  Although I haven't
			 run into a problem with it so far.  *shrug*.  However, I
			 think it only matters if the destination buffer's pointer
			 exists after the source's pointer.  But in our case it is
			 before so we should be safe because they will never cross
			 while incrementing during the copy.
			----------------------------------------------------------*/
			if( tmpBuf != buffer )
			{
				_tcscpy(buffer, tmpBuf);
			}
		}
#if 0 /* Don't do this, we want to return the buffer unmodified */
		else
		{
			/*----------------------------------------------------------
			 Everything was whitespace, so just terminate their return
			 buffer.  We still return FALSE even though we successfully
			 removed all whitespace, because we want to tell them that
			 there is no data in the returned buffer.
			----------------------------------------------------------*/
			*buffer = 0;
		}
#endif
	}

	return( 0 != length );
}


/*--------PL_SimpleToken------------------------------------------------
 Purpose:
 - To parse a string that is delimited by spaces or quotation marks. And
   it returns each token that is delimited by such, with the remainder
   of the string in destArgs.
----------------------------------------------------------------------*/
BOOL WINAPI PL_SimpleToken(
	 LPTSTR destBuf
	,LPTSTR destArgs
	,LPCTSTR srcBuf
)
{
	BOOL retVal = FALSE;

	/*------------------------------------------------------------------
	 No reason to let things crash, just cause of a buggy module,
	 so let's slow things down and play it safe.
	------------------------------------------------------------------*/
	if(
		NULL != srcBuf && 0 != *srcBuf && PLS_MAX_SIZE > _tcslen(srcBuf)
		&&
		NULL != destBuf
	)
	{
		_TCHAR tmpBuf[PLS_MAX_SIZE];

		/*--------------------------------------------------------------
		 So, lets make sure we are actually playing with a useful buffer
		--------------------------------------------------------------*/
		if( PL_StripWS(_tcscpy(tmpBuf, srcBuf)) )
		{
			UINT i = 0;

			/*----------------------------------------------------------
			 Now lets find our first token delimiter, checking that we
			 haven't reached the end of the string.

			 NOTE:
			 For now we allow double and single quotes and spaces. I'm
			 thinking we should remove support for single quotes however
			----------------------------------------------------------*/
			while(
				0        != tmpBuf[i]
				&&
				_T('\"') != tmpBuf[i]
				&&
				_T('\'') != tmpBuf[i]
				&&
				!_istspace( tmpBuf[i] )
			)
			{
				++i;
			}

			/*----------------------------------------------------------
			 Well, if we reached the end of the string, then return that
			 as the token, and be done with it.
			----------------------------------------------------------*/
			if( 0 == tmpBuf[i] )
			{
				_tcscpy(destBuf, tmpBuf);

				if( NULL != destArgs )
				{
					*destArgs = 0;
				}
			}
			/*----------------------------------------------------------
			 Ok, so we found a delimiter, was it the first character of
			 the string?  If so, lets go with that and parse it out
			 using the delimiter (which is the first char of tmpBuf[])
			 to find out where this token ends.
			----------------------------------------------------------*/
			else if( 0 == i )
			{
				_TCHAR delim;

				/*------------------------------------------------------
				 Ok, we gotta store our delimiter.
				------------------------------------------------------*/
				delim = *tmpBuf;

				/*------------------------------------------------------
				 Ok, lets look for the closing delimiter now.
				------------------------------------------------------*/
				while( delim != tmpBuf[++i] && 0 != tmpBuf[i] );

				/*------------------------------------------------------
				 Did we get a poorly formatted string? Well that is ok
				 we will just give them everything to the first
				 whitespace, or the end of the string if there is no WS.

				 NOTE: We could remove this, and just return everything
				 after the first delim to the end of the line.  But I
				 think this is ok for now.
				------------------------------------------------------*/
				if( 0 == tmpBuf[i] )
				{
					i = 0;

					/*--------------------------------------------------
					 ladeeeda, looking for whitespace... *yawn*
					--------------------------------------------------*/
					while( !_istspace(tmpBuf[++i]) && 0 != tmpBuf[i] );
				}

				/*------------------------------------------------------
				 So, we have a token now.  Cool deal, so let us
				 pass the parsed token back to them.

				 Remember, our token starts at tmpBuf+1, because
				 *tmpBuf itself was the start delim.

				 Also, since 'i' includes the delim characters in its
				 count, we need to use 'i-1', then put a NULL terminator
				 where the closing delim was.
				------------------------------------------------------*/
				_tcsncpy(destBuf, tmpBuf+1, i-1);
				destBuf[i-1] = 0;

				/*------------------------------------------------------
				 Do they want the remainder of the string passed back as
				 arguments?  That would only make sense, but if not,
				 then no worries, it goes to the bit bucket.  But if
				 they do, we'll make sure that its all nice and cleaned
				 up from any padded whitespace.
				------------------------------------------------------*/
				if( NULL != destArgs )
				{
					/*--------------------------------------------------
					 Check to see if there is anything remainding after
					 the first token.  It could be that the string ended
					 after the closing delim, so check the next char
					 after that as well.
					--------------------------------------------------*/
					if( 0 != tmpBuf[i] && 0 != tmpBuf[i+1] )
					{
						PL_StripWS(_tcscpy(destArgs, &tmpBuf[i+1]));
					}
					else
					{
						*destArgs = 0;
					}
				}
			}
			/*----------------------------------------------------------
			 Well, we did find a valid delim, however it was part way
			 through the string buffer. So, no need to really parse
			 anything more, because everything up to the delimiter is
			 actually the first token. Let us copy that over and then
			 once again, if they want the remainder returned as their
			 arguments, we will comply.
			----------------------------------------------------------*/
			else
			{
				_tcsncpy(destBuf, tmpBuf, i);
				destBuf[i] = 0;

				if( NULL != destArgs )
				{
					PL_StripWS(_tcscpy(destArgs, tmpBuf+i));
				}
			}

			/*----------------------------------------------------------
			 There is at least something in the destBuf now.  Maybe not
			 a whole lot, but what do we know... something is something
			 so give 'em a thumbs up.
			----------------------------------------------------------*/
			retVal = TRUE;
		}
	}

	return( retVal );
}


/*--------PL_BreakTokens-----------------------------------------------
 Purpose:
 - Breaks apart tokens seperated by brackets [ ] and puts everything 
   after the bracket set into the extra buffer.
 Notes:
 - A bracket set can be anywhere in the source string.  With one
   exception:
   If a bracket set is within a quotation set, the bracket set is
   ignored.  This fixes problems with nested brackets, used for other
   commands.  Example:
   LabelTextOrg "[cpu]"
   !execute [!LabelSetText $LabelTextOrg$][!LabelShow]
   While !execute is called, !execute's parameters will be expanded into
   [!LabelSetText "[cpu]"][!LabelShow]
   Before this fix, this function would have returned:
   !LabelSetText "[cpu
   Now it returns the expected
   !LabelSetText "[cpu]"

   Another option would have been to support nested brackets, and only
   break apart the outmost set of brackets.  However, for now I prefer
   the ignoring of content within quotation marks.  Although both ways
   are implemented, the nested support is #ifdef'd out

 - One typically calls this using the following code snippet:
 
   LPCTSTR pszString = _T("some [string] with [tokens][in it] ok");
   _TCHAR srcBuffer[PLS_MAX_SIZE];
   _TCHAR destBuffer[PLS_MAX_SIZE];

   _tcscpy(srcBuffer, pszString);

   while( PL_BreakTokens(destBuffer, srcBuffer, srcBuffer) )
   {
     // Do something with destBuffer
   }

   Thus, every time this function is called in the loop, srcBuffer will
   always have the remainder of the string being parsed, until there are
   no more bracket sets to be parsed out.  From the above example, we
   would have iterated 3 times successfully through the loop with
   destBuffer each time respectively containing (without the quotes):
   "string"
   "tokens"
   "in it"
----------------------------------------------------------------------*/
BOOL WINAPI PL_BreakTokens(
	 LPTSTR destBuf
	,LPTSTR extra
	,LPCTSTR srcBuf
)
{
	BOOL retVal = FALSE;

	/*------------------------------------------------------------------
	 No reason to let things crash, just cause of a buggy module,
	 so let's slow things down and play it safe.
	------------------------------------------------------------------*/
	if(
		NULL != srcBuf && 0 != *srcBuf && PLS_MAX_SIZE > _tcslen(srcBuf)
		&&
		NULL != extra
		&&
		NULL != destBuf
	)
	{
		LPCTSTR ptr1, ptr2;

#if 1 /* Support for ignoring brackets within quotation sets */

		/* No reason to assign ptr2 here, but gets rid of compiler warning */
		ptr1 = ptr2 = srcBuf;

		/*--------------------------------------------------------------
		 Now lets find our first open bracket, or ignore parts of the
		 string if it is within quotation marks.
		--------------------------------------------------------------*/
		while(
			0        != *ptr1
			&&
			_T('\"') != *ptr1
			&&
			_T('[')  != *ptr1
		)
		{
			++ptr1;
		}

		if(_T('\"') == *ptr1)
		{
			++ptr1;

			while(0 != *ptr1 && _T('\"') != *ptr1)
			{
				++ptr1;
			}

			if(0 != *ptr1)
			{
				/* Recursively call ourselves with the portion we haven't ignored */
				return(PL_BreakTokens(destBuf, extra, ptr1+1));
			}
		}
		/*--------------------------------------------------------------
		 If we found our opening bracket, then search for our closing
		 bracket, while ignoring anything in quotation sets.
		--------------------------------------------------------------*/
		else if(_T('[') == *ptr1)
		{
			ptr2 = ptr1;

			do
			{
				/* We are currently on some delim, increment past it */
				ptr2++;

				/* Find next delim */
				while(
					0        != *ptr2
					&&
					_T('\"') != *ptr2
					&&
					_T(']')  != *ptr2
				)
				{
					++ptr2;
				}

				if(_T('\"') == *ptr2)
				{
					++ptr2;

					while(0 != *ptr2 && _T('\"') != *ptr2)
					{
						++ptr2;
					}
				}
			} while(0 != *ptr2 && _T(']') != *ptr2);
		}

		if(_T('[') == *ptr1 && _T(']') == *ptr2)

#elif 0 /* This adds support for nested brackets */

		LPCTSTR pItr;
		UINT nNested;

		nNested = 0;
		ptr1 = ptr2 = NULL;
		pItr = srcBuf;

		while(0 != *pItr && (NULL == ptr1 || NULL == ptr2))
		{
			if(_T('[') == *pItr)
			{
				if(0 == nNested)
				{
					ptr1 = pItr;
					ptr2 = NULL;
				}

				nNested++;
			}
			else if(0 != nNested && _T(']') == *pItr)
			{
				nNested--;

				if(0 == nNested)
				{
					ptr2 = pItr;
				}
			}

			++pItr;
		}

		if(
			NULL != ptr1 && _T('[') == *ptr1
			&&
			NULL != ptr2 && _T(']') == *ptr2
		)

#else /* Old school style.  Simple, but not very effective */

		/*--------------------------------------------------------------
		 Alright, so is there a valid bracket set in their src string?
		--------------------------------------------------------------*/
		if(
			NULL != (ptr1 = _tcschr(srcBuf, _T('[')))
			&&
			NULL != (ptr2 = _tcschr(ptr1, _T(']')))
		)
#endif
		{
			_TCHAR tmpBuf[PLS_MAX_SIZE];

			/* Increment past the opening delimiter */
			ptr1++;

			/*----------------------------------------------------------
			 Lets copy what we want into a working buffer
			----------------------------------------------------------*/
			_tcsncpy(tmpBuf, ptr1, ptr2-ptr1);
			*(tmpBuf + (ptr2 - ptr1)) = 0;

			/*----------------------------------------------------------
			 Clean it up. Then put it into their destination buffer

			 NOTE: By doing this, it doesn't allow for whitespace to be
			 passed back if it was inside the bracket set.  Is this OK?
			 ATM we only ever preserve whitespace inside of quotes.
			----------------------------------------------------------*/
			*destBuf = 0;

			if( PL_StripWS(tmpBuf) )
			{
				_tcscpy(destBuf, tmpBuf);
			}

			/* Move the pointer past the closing delimiter.*/
			ptr2++;
			*extra = 0;

			/*----------------------------------------------------------
			 Clean up the remainder and put it into the extra buffer.
			----------------------------------------------------------*/
			if( PL_StripWS(_tcscpy(tmpBuf, ptr2)) )
			{
				_tcscpy(extra, tmpBuf);
			}

			/*----------------------------------------------------------
			 Give 'em the thumbs up, even though it could have been an
			 empty bracket set.  There was still a set there, and so we
			 say yeehaw to it.
			----------------------------------------------------------*/
			retVal = TRUE;
		}
	}

	return( retVal );
}

/*----------------------------------------------------------------------
 END OF FILE
----------------------------------------------------------------------*/
